//
//  ColumnTextView.m
//  TextKit
//
//  Created by Gene Whitaker on 8/5/15.
//  Copyright (c) 2015 Bombing Brain Interactive. All rights reserved.
//

#import "ColumnTextView.h"
#import "Constants.h"
#import "ColumnTextContainer.h"
#import "CalloutView.h"

@interface ColumnTextView () <NSLayoutManagerDelegate, CalloutViewDelegate>
@property (strong, nonatomic) NSTextStorage *textStorage;
@property (strong, nonatomic) NSLayoutManager *layoutManager;
@property (nonatomic) NSInteger numberOfColumns;
@property (strong, nonatomic) NSString *documentPath;
@property (nonatomic) NSUInteger currentLineSpacingStartingIndex;
@property (strong, nonatomic) NSMutableArray *calloutViews;
@property (nonatomic) NSRange selectedRange;
@end

@implementation ColumnTextView

#pragma mark - view lifecycle

- (void)awakeFromNib {
    
    [super awakeFromNib];
    
    //Get the article text
    NSURL *documentsURL = [[[NSFileManager defaultManager] URLsForDirectory:NSDocumentDirectory inDomains:NSUserDomainMask] lastObject];
    self.documentPath = [documentsURL.path stringByAppendingPathComponent:@"Article.data"];
    
    //Create and load the text storage
    self.textStorage = [[NSTextStorage alloc] initWithAttributedString:[NSKeyedUnarchiver unarchiveObjectWithFile:self.documentPath]];
    
    //Create and set up the layout manager
    self.layoutManager = [[NSLayoutManager alloc] init];
    self.layoutManager.delegate = self;
    self.layoutManager.allowsNonContiguousLayout = NO;
    
    //Add the layout manager to the text storage
    [self.textStorage addLayoutManager:self.layoutManager];
    
    //Automatically redraw the view on bounds change (forces text layout)
    self.contentMode = UIViewContentModeRedraw;

    //Set up an index to keep track of line spacing layout
    self.currentLineSpacingStartingIndex = 0;
    
    //Create an array to store references to callout views
    self.calloutViews = [[NSMutableArray alloc] init];
    
    //Set up a range to track currently selected text for audio playback
    self.selectedRange = NSMakeRange(0,0);
    
    //Set up hit-test gesture
    UITapGestureRecognizer *tap = [[UITapGestureRecognizer alloc] initWithTarget:self action:@selector(tap:)];
    [self addGestureRecognizer:tap];

}


#pragma mark - Drawing


- (void)drawRect:(CGRect)rect {
    
    //Remove existing callout views
    [self.calloutViews makeObjectsPerformSelector:@selector(removeFromSuperview)];
    [self.calloutViews removeAllObjects];
    
    //Remove existing text containers (they will be redrawn)
    for (NSUInteger i=self.layoutManager.textContainers.count; i>0; i--) {
        [self.layoutManager removeTextContainerAtIndex:i-1];
    }
    
    //Determine the number of columns based on aspect ratio of the view
    self.numberOfColumns = rect.size.height > rect.size.width ? 2 : 3;
    
    //Determine the size of the text containers
    CGSize containerSize = CGSizeMake(rect.size.width/self.numberOfColumns, rect.size.height);
    
    for (int i=0; i<self.numberOfColumns; i++) {
        
        //Determine the rect of the column so we can place exclusion paths in the right container
        CGRect containerRect = CGRectMake(i*containerSize.width, 0, containerSize.width, containerSize.height);
        
        //Create the container
        ColumnTextContainer *container = [[ColumnTextContainer alloc] initWithSize:containerRect.size];

        //Look for eclusion paths intersecting for this container and add the intersecting area
        NSMutableArray *containerExclusionPaths = [[NSMutableArray alloc] init];
       for (UIBezierPath *exclusionPath in self.exclusionPaths) {
            
            if (CGRectIntersectsRect(containerRect, exclusionPath.bounds)) {
                
                //Get the intesecting rect
                CGRect intersectingRect = CGRectIntersection(containerRect, exclusionPath.bounds);
                
                //Convert the intersecting rect to the container's coordinate system
                CGRect intersectingRectInContainerCoordinates = CGRectMake(intersectingRect.origin.x - containerRect.origin.x,
                                                                           intersectingRect.origin.y,
                                                                           intersectingRect.size.width,
                                                                           intersectingRect.size.height);
                
                //add the intesecting path to the container's exclusion path array
                UIBezierPath *intersectingPath = [UIBezierPath bezierPathWithRect:intersectingRectInContainerCoordinates];
                [containerExclusionPaths addObject:intersectingPath];
                
            }
            
        }
        //Add the exclusion paths to the container
        container.exclusionPaths = containerExclusionPaths;
        
        //Capture the frame of this container (needed for when we actually layout the text)
        container.frame = containerRect;
        
        //Capture the column index for this container
        container.columnIndex = i;
        
        //Add the text container to the layoutmanager's array of containers
        [self.layoutManager addTextContainer:container];
        
    }
    
    //Draw the text in the containers
    [self layoutText];
    
}


-(void)layoutText {
    
    //Loop through the text containers and draw
    for (ColumnTextContainer *container in self.layoutManager.textContainers) {
        
        //Tell the layout manager to draw the text for this container
        NSRange glyphRange = [self.layoutManager glyphRangeForTextContainer:container];
        [self.layoutManager drawGlyphsForGlyphRange:glyphRange atPoint:container.frame.origin];
        
    }

}


#pragma mark - Gestures


-(void)tap:(UITapGestureRecognizer*)tap {

    //Get the location of the touch
    CGPoint touchPoint = [tap locationInView:self];
    
    //Identify the text container for the touch (if applicable)
    ColumnTextContainer *container = [self textContainerForPoint:touchPoint];
    if (container) {
        
        //Get the character that was touched in this text container
        NSUInteger charIndex = [self.layoutManager
                                characterIndexForPoint:[container pointInTextContainerCoordinates:touchPoint]
                                inTextContainer:container
                                fractionOfDistanceBetweenInsertionPoints:nil];
        
        //Look for AUDIO_FILE_ATTRIBUTE on this character,
        //and get the full effective range for this attribute if so
        NSRange effectiveRange;
        NSString *audioAttribute = [self.textStorage attribute:AUDIO_FILE_ATTRIBUTE atIndex:charIndex effectiveRange:&effectiveRange];
        
        //If AUDIO_FILE_ATTRIBUTE found, process
        if (audioAttribute) {
            if (effectiveRange.location == self.selectedRange.location) {
                
                //If this range was previously tapped, reset
                self.selectedRange = NSMakeRange(0,0);
                
                //stop audio
                if (self.delegate && [self.delegate respondsToSelector:@selector(columnTextViewDidToggleAudio:on:)]) {
                    [self.delegate columnTextViewDidToggleAudio:audioAttribute on:NO];
                }
            }
            else {
                
                //If this is a new range, set the current selected range property
                self.selectedRange = effectiveRange;
                
                //Start audio
                if (self.delegate && [self.delegate respondsToSelector:@selector(columnTextViewDidToggleAudio:on:)]) {
                    [self.delegate columnTextViewDidToggleAudio:audioAttribute on:YES];
                } 
            }
            
            //Force a redraw of the view so that line spacing can be adjusted and audio elements displayed
            [self setNeedsDisplay];
            
        }
    }
}

-(ColumnTextContainer*)textContainerForPoint:(CGPoint)point {
    //Determine which text container contains the given point
    int i=0;
    for (ColumnTextContainer *container in self.layoutManager.textContainers) {
        if (CGRectContainsPoint(container.frame, point)) {
            NSLog(@"container tap: %d",i);
            return container;
        }
        i++;
    }
    return nil;
}


#pragma mark - NSLayoutManagerDelegate methods


- (CGFloat)layoutManager:(NSLayoutManager *)layoutManager lineSpacingAfterGlyphAtIndex:(NSUInteger)glyphIndex withProposedLineFragmentRect:(CGRect)rect {
    
    //Get the corresponding character index for the glyph ending the line
    NSUInteger characterIndex = [layoutManager characterIndexForGlyphAtIndex:glyphIndex];
    
    //Get the range of characters on this line
    NSRange range = NSMakeRange(self.currentLineSpacingStartingIndex, characterIndex-self.currentLineSpacingStartingIndex);
    
    //Determine if self.selectedRange intersects with this line
    BOOL hasCallout = (NSIntersectionRange(range, self.selectedRange).length > 0);
    
    //Set the next starting index
    if (characterIndex != self.textStorage.length-1) {
        self.currentLineSpacingStartingIndex = characterIndex+1;
    }
    
    //If this line has an AUDIO_FILE_ATTRIBUTE, add additional line spacing for it
    if (hasCallout) {
        return CALLOUT_HEIGHT;
    }
    
    //Default line spacing
    return 0;
    
}

-(void)layoutManagerDidInvalidateLayout:(NSLayoutManager *)sender {
    
    //Reset our line spacing index every time the layout is invalidated
    self.currentLineSpacingStartingIndex = 0;
    
}

-(void)layoutManager:(NSLayoutManager *)layoutManager didCompleteLayoutForTextContainer:(ColumnTextContainer *)textContainer atEnd:(BOOL)layoutFinishedFlag {
    
//    NSLog(@"Column:%@ Index: %ld, Layout Finished: %d",textContainer.description,(long)textContainer.columnIndex,layoutFinishedFlag);
    
    //If a valid textContainer is returned, process layout
    if(textContainer) {
        
        //Get the glyph range for this container
        NSRange glyphRange = [self.layoutManager glyphRangeForTextContainer:textContainer];
        
        //Get the corresponding character range
        NSRange characterRange = [self.layoutManager characterRangeForGlyphRange:glyphRange actualGlyphRange:nil];
        
        //Does the characterRange intersect thr currently selected range
        BOOL hasCallout = (NSIntersectionRange(characterRange, self.selectedRange).length > 0);
        
        //If YES, position an audio callout
        if (hasCallout) {

            //Get the location of the text with this attribute
            CGRect rect = [self.layoutManager boundingRectForGlyphRange:self.selectedRange inTextContainer:textContainer];
             
            //Determine the location of the callout
            CGRect calloutRect = CGRectMake(rect.origin.x, rect.origin.y+rect.size.height-CALLOUT_HEIGHT,
                                               CallOUT_WIDTH, CALLOUT_HEIGHT);

            //Convert to view coordinates
            CGPoint calloutOriginInViewCoordinates = CGPointMake(textContainer.frame.origin.x+calloutRect.origin.x, textContainer.frame.origin.y+calloutRect.origin.y);
            CGRect calloutRectInViewCoordinates = CGRectMake(calloutOriginInViewCoordinates.x, calloutOriginInViewCoordinates.y, calloutRect.size.width, calloutRect.size.height);

            //Create and add the callout view to the text view
            CalloutView *calloutView = [[CalloutView alloc] initWithFrame:calloutRectInViewCoordinates];
            calloutView.delegate = self;
            calloutView.range = self.selectedRange;
            calloutView.allowsSelectedState = YES;
            calloutView.selected = YES;
            [self.calloutViews addObject:calloutView];
            [self addSubview:calloutView];
            
         }

    }
}


#pragma mark - CalloutViewDelegate

-(void)calloutViewTapped:(CalloutView*)calloutView fromRect:(CGRect)rect withRange:(NSRange)range {
    
    //Find the AUDIO_FILE_ATTRIBUTE so we know what audio to play
    NSDictionary *attributes = [self.textStorage attributesAtIndex:range.location effectiveRange:nil];
    NSString *audioAttribute = attributes[AUDIO_FILE_ATTRIBUTE];
    
    //Play audio if found
    if (audioAttribute) {
        if (self.delegate && [self.delegate respondsToSelector:@selector(columnTextViewDidToggleAudio:on:)]) {
            [self.delegate columnTextViewDidToggleAudio:audioAttribute on:calloutView.selected];
        }
    }
    

}

@end
