//
//  EditorViewController.m
//  TextKit
//
//  Created by Gene Whitaker on 8/5/15.
//  Copyright (c) 2015 Bombing Brain Interactive. All rights reserved.
//

#import "EditorViewController.h"
#import "Constants.h"
#import "CustomTextView.h"
#import "CustomTextStorage.h"
#import "AudioFileTableViewController.h"
#import "CalloutView.h"

@interface EditorViewController () <NSLayoutManagerDelegate, AudioFileTableViewControllerDelegate, CalloutViewDelegate>
@property (weak, nonatomic) IBOutlet CustomTextView *textView;
@property (strong, nonatomic) CustomTextStorage *textStorage;
@property (strong, nonatomic) NSString *documentPath;
@property (nonatomic) NSUInteger currentLineSpacingStartingIndex;
@property (strong, nonatomic) NSMutableArray *calloutViews;
@end

@implementation EditorViewController

#pragma mark - view lifecycle

- (void)viewDidLoad {
    
    [super viewDidLoad];
    
    //Get the document path for the text
    NSURL *documentsURL = [[[NSFileManager defaultManager] URLsForDirectory:NSDocumentDirectory inDomains:NSUserDomainMask] lastObject];
    self.documentPath = [documentsURL.path stringByAppendingPathComponent:@"Article.data"];
    
    //Set up the text view
    self.textView.allowsEditingTextAttributes = YES;
    
    //Set up custom text storage
    self.textStorage = [[CustomTextStorage alloc] init];
    
    //Add the text view's layoutmanager to the custom text storage
    [self.textStorage  addLayoutManager: self.textView.layoutManager];
    
    //Set up Layout Manager
    self.textView.layoutManager.delegate = self;
    self.textView.layoutManager.allowsNonContiguousLayout = NO;
    
    //Set up an index to keep track of line spacing layout
    self.currentLineSpacingStartingIndex = 0;
    
    //Create an array to store references to callout views
    self.calloutViews = [[NSMutableArray alloc] init];
    
    //Add a menu option to add the AUDIO_FILE_ATTRIBUTE custom text attribute
    UIMenuItem *menuItem = [[UIMenuItem alloc] initWithTitle:@"Insert Audio..." action:@selector(attachAudio:)];
    [[UIMenuController sharedMenuController] setMenuItems:[NSArray arrayWithObject:menuItem]];

}

-(void)viewWillDisappear:(BOOL)animated {
    
    //Save the text to file on exit
    [NSKeyedArchiver archiveRootObject:self.textStorage toFile:self.documentPath];
    
}

-(void)viewDidAppear:(BOOL)animated {
    
    //Late-load the text so that we are sure final textview geometry has been computed before text layout
    [self.textStorage appendAttributedString:[NSKeyedUnarchiver unarchiveObjectWithFile:self.documentPath]];
    
}

#pragma mark - implementation

-(void)attachAudio:(id)sender {
    
    //Show an audio file selector, anchored to the selected text
    AudioFileTableViewController *audioFileSelector = [[AudioFileTableViewController alloc] init];
    audioFileSelector.delegate = self;
    audioFileSelector.modalPresentationStyle = UIModalPresentationPopover;
    UIPopoverPresentationController *popPC = audioFileSelector.popoverPresentationController;
    popPC.sourceView = self.textView;
    popPC.sourceRect = [self.textView.layoutManager boundingRectForGlyphRange:self.textView.selectedRange inTextContainer:self.textView.textContainer];
    [self presentViewController:audioFileSelector animated:YES completion:nil];
    
}

-(void)positionCallouts {
    
    if (self.textStorage.length) {
        
        //Remove existing callout views
        [self.calloutViews makeObjectsPerformSelector:@selector(removeFromSuperview)];
        [self.calloutViews removeAllObjects];
        
        //Enumerate the attributes in the text
        [self.textStorage enumerateAttributesInRange:NSMakeRange(0, self.textStorage.length) options:NSAttributedStringEnumerationLongestEffectiveRangeNotRequired usingBlock:
         ^(NSDictionary *attributes, NSRange range, BOOL *stop) {
             
             //Get the AUDIO_FILE_ATTRIBUTE value
             NSString *audioAttribute = attributes[AUDIO_FILE_ATTRIBUTE];
             
             if (audioAttribute) {
                 
                 //Get the bounding rect of the text with this attribute
                 CGRect rect = [self.textView.layoutManager boundingRectForGlyphRange:range inTextContainer:self.textView.textContainer];
                 
                 //Determine the location of the callout (below the text)
                 CGPoint calloutPoint = CGPointMake(rect.origin.x, rect.origin.y+rect.size.height-CALLOUT_HEIGHT);
                 
                 //Adjust for UITextView inset
                 CGRect calloutRect = CGRectMake(calloutPoint.x, calloutPoint.y+self.textView.textContainerInset.top, CallOUT_WIDTH, CALLOUT_HEIGHT);
                 
                 //Create the callout view
                 CalloutView *calloutView = [[CalloutView alloc] initWithFrame:calloutRect];
                 calloutView.delegate = self;
                 
                 //Callout view remembers the range it refers to
                 calloutView.range = range;

                 //Add the callout to the textview
                 [self.calloutViews addObject:calloutView];
                 [self.textView addSubview:calloutView];
             }
         }];
    }
    
}

#pragma mark - NSLayoutManagerDelegate

- (CGFloat)layoutManager:(NSLayoutManager *)layoutManager lineSpacingAfterGlyphAtIndex:(NSUInteger)glyphIndex withProposedLineFragmentRect:(CGRect)rect {

    //Get the corresponding character index for the glyph ending the line
    NSUInteger characterIndex = [layoutManager characterIndexForGlyphAtIndex:glyphIndex];
        
    //Get the range of characters on this line
    NSRange range = NSMakeRange(self.currentLineSpacingStartingIndex, characterIndex-self.currentLineSpacingStartingIndex);
        
    //Set the next starting index
    if (characterIndex != self.textStorage.length-1) {
        self.currentLineSpacingStartingIndex = characterIndex+1;
    }
    
    //Enumerate the attributes on this line
    __block BOOL hasCallout = NO;
    [self.textStorage enumerateAttributesInRange:range options:NSAttributedStringEnumerationLongestEffectiveRangeNotRequired usingBlock:
     ^(NSDictionary *attributes, NSRange range, BOOL *stop) {
         
         //Determine if AUDIO_FILE_ATTRIBUTE exists on this line of text
         NSString *audioAttribute = attributes[AUDIO_FILE_ATTRIBUTE];
         if (audioAttribute) {
             hasCallout = YES;
             *stop = YES;
         }
     }];
    
    //If this line has an AUDIO_FILE_ATTRIBUTE, add additional line spacing for it
    if (hasCallout) {
        return CALLOUT_HEIGHT;
    }
    
    //Default line spacing
    return 0;
    
}

-(void)layoutManager:(NSLayoutManager *)layoutManager textContainer:(NSTextContainer *)textContainer didChangeGeometryFromSize:(CGSize)oldSize {
    
    //Reset our line spacing index every time the layout is invalidated or geometry changed
    self.currentLineSpacingStartingIndex = 0;

}

-(void)layoutManagerDidInvalidateLayout:(NSLayoutManager *)sender {
    
    //Reset our line spacing index every time the layout is invalidated
    self.currentLineSpacingStartingIndex = 0;
    
}

-(void)layoutManager:(NSLayoutManager *)layoutManager didCompleteLayoutForTextContainer:(NSTextContainer *)textContainer atEnd:(BOOL)layoutFinishedFlag {

    //Layout callout views
    [self positionCallouts];
    
}

#pragma mark - AudioFileTableViewControllerDelegate

-(void)audioFileTableViewControllerDidSelectFile:(NSString*)audioFileName {
    
    [self dismissViewControllerAnimated:YES completion:nil];
    
    //Copy original text attributes dictionary at the starting index
    NSDictionary *originalAttributes = [self.textStorage attributesAtIndex:self.textView.selectedRange.location effectiveRange:nil];
    NSMutableDictionary *attributes = [originalAttributes mutableCopy];

    if (audioFileName.length) {
        
        //If an audio file was selected, add a custom AUDIO_FILE_ATTRIBUTE to the text
        attributes[AUDIO_FILE_ATTRIBUTE] = audioFileName;
        
        //Also set the selected text's traits to bold, italic
        UIFont *oldFont = attributes[NSFontAttributeName];
        UIFontDescriptor *oldFontDescriptor = [oldFont fontDescriptor];
        UIFontDescriptor *italicDescriptor = [oldFontDescriptor fontDescriptorWithSymbolicTraits:(oldFontDescriptor.symbolicTraits | UIFontDescriptorTraitItalic)];
        UIFontDescriptor *boldDescriptor = [italicDescriptor fontDescriptorWithSymbolicTraits:(italicDescriptor.symbolicTraits | UIFontDescriptorTraitBold)];
        UIFont *newFont = [UIFont fontWithDescriptor:boldDescriptor size:oldFont.pointSize];
        if(newFont != nil) {
            //Add the bold italic font attribute
            attributes[NSFontAttributeName] = newFont;
        }
        
    }
    else {
        
        //If the audio file name is empty, remove the custom ATTRIBUTE
        [attributes removeObjectForKey:AUDIO_FILE_ATTRIBUTE];
        
        //Also remove the text's bold, italic traits
        UIFont *oldFont = attributes[NSFontAttributeName];
        UIFontDescriptor *oldFontDescriptor = [oldFont fontDescriptor];
        UIFontDescriptor *italicDescriptor = [oldFontDescriptor fontDescriptorWithSymbolicTraits:(oldFontDescriptor.symbolicTraits & ~UIFontDescriptorTraitItalic)];
        UIFontDescriptor *boldDescriptor = [italicDescriptor fontDescriptorWithSymbolicTraits:(italicDescriptor.symbolicTraits & ~UIFontDescriptorTraitBold)];
        UIFont *newFont = [UIFont fontWithDescriptor:boldDescriptor size:oldFont.pointSize];
        if(newFont != nil) {
            attributes[NSFontAttributeName] = newFont;
        }
        
    }
    
    //Set the attributes on the selected text
    [self.textStorage setAttributes:attributes range:self.textView.selectedRange];
    [self.textView resignFirstResponder];

}

#pragma mark - CalloutViewDelegate

-(void)calloutViewTapped:(id)calloutView fromRect:(CGRect)rect withRange:(NSRange)range {

    //Select the range attached this callout
    [self.textView setSelectedRange:range];
    
    //show the audio options popover
    [self attachAudio:nil];
    
}

@end
